/*
 * $Id: ataprint.c 631969 2014-02-26 04:09:54Z nrupesh $
 *
 * Copyright (C) 2000 Andre Hedrick <andre@linux-ide.org>
 * Copyright (C) 1999-2000 Michael Cornwell <cornwell@acm.org>
 *
 * Copyright (c) 2002-2012, Juniper Networks, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * You should have received a copy of the GNU General Public License
 * (for example COPYING); if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

  
#include <sys/types.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/jnx/dkio.h>
#include <sys/ata.h>
#include <sys/fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <sys/wait.h>

#include "atacmds.h"
#include "smartd.h"
#include "extern.h"
#include "atavendor_db.h"


static void ataprintdriveinfo (struct ata_params *);
static void ataprintgeneralsmartvalues (struct ata_smart_values);
static void __unused ataprintsmartthresholds (struct ata_smart_thresholds);
static void ataprintsmarterrorlog ( struct ata_smart_errorlog);
static void printsmartattribwiththres ( struct ata_smart_values data, 
				struct ata_smart_thresholds thresholds);
static void ataprintsmartselftestlog (struct ata_smart_selftestlog data);
static void atapsuedochecksmart ( struct ata_smart_values , 
				struct ata_smart_thresholds );
static void printsmartofflinestatus (struct ata_smart_values);
static void printsmartselfexecstatus (struct ata_smart_values);
static void printsmarttotaltimecompleteoffline (struct ata_smart_values);
static void printsmartofflinecollectcap (struct ata_smart_values);
static void printsmartcapability (struct ata_smart_values);
static void printsmarterrorlogcapability (struct ata_smart_values);
static void printsmartshortselftestpollingtime (struct ata_smart_values);
static void printsmartextendedselftestpollingtime (struct ata_smart_values);
void atacaptivefinished ( int signo);

void 
ataprintdriveinfo (struct ata_params *drive)
{
    unsigned int i;

    printf ("Device: %s  Supports ATA Version %i, Firmware version ", 
		drive->model, ataversioninfo (drive));
    for (i=0; i<sizeof(drive->revision) && drive->revision[i]; i++)
	printf("%c", drive->revision[i]);
    printf("\n");

    cap_print(drive);
    /* JUNOS */
    atavendorattributesdbupdate(drive->model);
}

/* 
 * Prints verbose Off-line data collection status byte
 */
static void 
printsmartofflinestatus (struct ata_smart_values data)
{
    printf ("Off-line data collection status: ");    
    switch (data.offline_data_collection_status) {
    case  0x0: 
    case 0x80:
        printf ("(0x%02x)\tOffline data collection activity was\n\t\t\t\t\t",
        data.offline_data_collection_status);
        printf("never started\n");
        break;
    
    case  0x01: 
    case 0x81:
        printf ("(0x%02x)\tReserved\n",
        data.offline_data_collection_status);
        break;

    case  0x02: 
    case 0x82:
        printf ("(0x%02x)\tOffline data collection activity \n\t\t\t\t\t",
        data.offline_data_collection_status);
        printf ("completed without error\n");
        break;
    
    case  0x03: 
    case 0x83:
        printf ("(0x%02x)\tReserved\n",
        data.offline_data_collection_status);
        break;
        
    case  0x04: 
    case 0x84:
        printf ("(0x%02x)\tOffline data collection activity was \n\t\t\t\t\t",
        data.offline_data_collection_status);
        printf ("suspended by an interrupting command\n");
        break;
    
    case  0x05: 
    case 0x85:
        printf ("(0x%02x)\tOffline data collection activity was \n\t\t\t\t\t",
             data.offline_data_collection_status);
        printf ("aborted by an interrupting command\n");
        break;
    
    case  0x06: 
    case 0x86:
        printf ("(0x%02x)\tOffline data collection activity was \n\t\t\t\t\t",
        data.offline_data_collection_status);
        printf ("aborted by the device with fatal error\n");
        break;
    
    default:
        if (((data.offline_data_collection_status >= 0x07) &&
           (data.offline_data_collection_status <= 0x3f)) ||
             ((data.offline_data_collection_status >= 0xc0))) {
            printf ("(0x%02x)\tVendor Specific\n",
                    data.offline_data_collection_status);
        } else {
            printf ("(0x%02x)\tReserved\n",
                    data.offline_data_collection_status);
        }
    }
}

/* 
 * Prints verbose self-test status byte
 */
static void 
printsmartselfexecstatus (struct ata_smart_values data)
{
    printf ("Self-test execution status:      ");
    switch (data.self_test_exec_status >> 4) {
    case  0:
        printf ("(%4d)\tThe previous self-test routine completed\n\t\t\t\t\t",
            data.self_test_exec_status);
        printf ("without error or no self-test has ever \n\t\t\t\t\tbeen run\n");
        break;
    case  1:
        printf ("(%4d)\tThe self-test routine was aborted by\n\t\t\t\t\t",
            data.self_test_exec_status);
        printf ("the host\n");
        break;
    case  2:
        printf ("(%4d)\tThe self-test routine was interrupted\n\t\t\t\t\t",
            data.self_test_exec_status);
        printf ("by the host with a hard or soft reset\n");
        break;
    case  3:
        printf ("(%4d)\tA fatal error or unknown test error\n\t\t\t\t\t",
            data.self_test_exec_status);
        printf ("occurred while the device was executing\n\t\t\t\t\t");
        printf ("its self-test routine and the device \n\t\t\t\t\t");
        printf ("was unable to complete the self-test \n\t\t\t\t\t");
        printf ("routine\n");
        break;
    case  4:
        printf ("(%4d)\tThe previous self-test completed having\n\t\t\t\t\t",
            data.self_test_exec_status);
        printf ("a test element that failed and the test\n\t\t\t\t\t");
        printf ("element that failed is not known\n");
        break;
    case  5:
        printf ("(%4d)\tThe previous self-test completed having\n\t\t\t\t\t",
            data.self_test_exec_status);
        printf ("the electrical element of the test\n\t\t\t\t\t");
        printf ("failed\n");
        break;
    case  6:
        printf ("(%4d)\tThe previous self-test completed having\n\t\t\t\t\t",
            data.self_test_exec_status);
        printf ("the servo (and/or seek) element of the \n\t\t\t\t\t");
        printf ("test failed\n");
        break;
    case  7:
        printf ("(%4d)\tThe previous self-test completed having\n\t\t\t\t\t",
            data.self_test_exec_status);
        printf ("the read element of the test failed\n");
        break;
    case  15:
        printf ("(%4d)\tSelf-test routine in progress\n\t\t\t\t\t",
            data.self_test_exec_status);
        printf ("%1d0%% of test remaining\n", data.self_test_exec_status & 0x0f);
        break;
    default:
        printf ("(%4d)\tReserved\n", data.self_test_exec_status);
        break;
    }
}

/* 
 * Prints time to complete off-line data collection
 */
static void 
printsmarttotaltimecompleteoffline (struct ata_smart_values data)
{
    printf ("Total time to complete off-line \n");
    printf ("data collection: \t\t (%4d) Seconds\n", 
            data.total_time_to_complete_off_line);
}

/* 
 * Prints Off-line data collection capability
 */
static void 
printsmartofflinecollectcap (struct ata_smart_values data)
{
    printf ("Offline data collection \n");
    printf ("Capabilities: \t\t\t (0x%02x)",
        data.offline_data_collection_capability);

    if (data.offline_data_collection_capability == 0x00) {
        printf ("\tOff-line data collection not supported\n");
    } else {
        if (issupportexecuteofflineimmediate(data)) {
            printf (" SMART EXECUTE OFF-LINE IMMEDIATE\n");
        } else {
            printf ("NO SMART EXECUTE OFF-LINE IMMEDIATE\n");
        }

        if (issupportautomatictimer (data)) {
            printf ("\t\t\t\t\tAutomatic timer ON/OFF support\n");
        } else {
            printf ("\t\t\t\t\tNO Automatic timer ON/OFF support\n");
        }

        if (issupportofflineabort (data)) {
            printf ("\t\t\t\t\tAbort Offline Collection upon new\n");
            printf ("\t\t\t\t\tcommand\n");
        } else {
            printf ("\t\t\t\t\tSuspend Offline Collection upon new\n");
            printf ("\t\t\t\t\tcommand\n");
        }

        if (issupportofflinesurfacescan (data)) {
            printf ("\t\t\t\t\tOffline surface scan supported\n");
        } else {
            printf ("\t\t\t\t\tNO Offline surface scan supported\n");
        }

        if (issupportselftest(data)) {
	    printf ("\t\t\t\t\tSelf-test supported\n");
        } else {
	    printf ("\t\t\t\t\tNO Self-test supported\n");
        }
    }
}

/* 
 * Prints SMART capabilities
 */
static void 
printsmartcapability (struct ata_smart_values data)
{
    printf ("Smart Capablilities:       ");
    printf ("(0x%04x)\t", data.smart_capability);
    if (data.smart_capability == 0x00) {
        printf ("automatic saving of SMART data"); 
        printf ("\t\t\t\t\tis not implemented\n");
    } else {
        if (data.smart_capability & 0x01) {
	    printf ("Saves SMART data before entering \n");
	    printf ("\t\t\t\t\tpower-saving mode\n");
        } else {
	    printf ("does not save SMART data before  \n");
	    printf ("\t\t\t\t\tentering power-saving mode\n");
        }
        
        if (data.smart_capability & 0x02) {
            printf ("\t\t\t\t\tSupports SMART auto save timer\n");
        }
    }
}

/* 
 * Prints error log capabilities
 */
static void 
printsmarterrorlogcapability (struct ata_smart_values data)
{
    printf ("Error logging capability:       ");
    
    if (issmarterrorlogcapable(data)) {
        printf (" (0x%02x)\tError logging supported\n",
            data.errorlog_capability);
    } else {
        printf (" (0x%02x)\tError logging NOT supported\n",
            data.errorlog_capability);
    }
}

/* 
 * Prints short self test polling time
 */
static void 
printsmartshortselftestpollingtime (struct ata_smart_values data)
{
    if (issupportselftest(data)) {
        printf ("Short self-test routine \n");
        printf ("recommended polling time: \t (%4d) Minutes\n", 
            data.short_test_completion_time);
    } else {
        printf ("Short self-test routine \n");
        printf ("recommended polling time: \t    Not Supported\n");
    }
}

/* 
 * Prints extended self test polling time
 */
static void 
printsmartextendedselftestpollingtime (struct ata_smart_values data)
{
    if (issupportselftest(data)) {
        printf ("Extended self-test routine \n");
        printf ("recommended polling time: \t (%4d) Minutes\n", 
            data.extend_test_completion_time);
    } else {
        printf ("Extended self-test routine \n");
        printf ("recommended polling time: \t    Not Supported\n");
    }
}

/* 
 * Prints verbose SMART attributes and thresholds
 */
void 
printsmartattribwiththres (struct ata_smart_values data, struct ata_smart_thresholds thresholds)
{
    int i,j;
    printf ("Vendor Specific SMART Attributes with Thresholds:\n");
    printf ("Revision Number: %i\n", data.revnumber);
    printf ("     Attribute               Flag     Value Worst Threshold Raw Value\n");
    
    for (i = 0 ; i < NUMBER_ATA_SMART_ATTRIBUTES ; i++) {
	if ((data.vendor_attributes[i].id !=0) &&
	    (thresholds.thres_entries[i].id != 0)) {

	    printf("%s", atagetsmartattribname(data.vendor_attributes[i].id));
	    printf(" 0x%04x   %.3i   %.3i   %.3i       ", 
	    /* JUNOS begin
	     * The fields of the Unigen smart attribute structure are not same
	     * as the smart atrribute structure defined in by smartd. For
	     * Unigen SSD's the correct value of the fields are got by
	     * calling the function printsmartvalues.
	     */
	    (isunigenattributetbl() ?
	     printsmartvalues(&(data.vendor_attributes[i]), STATUS_FLAG, 0 ) :
	     data.vendor_attributes[i].status_flag),
	    (isunigenattributetbl() ?
	     printsmartvalues(&(data.vendor_attributes[i]), NORMAL_VALUE, 0 ) :
	     data.vendor_attributes[i].normalized),
	    (isunigenattributetbl() ?
	     printsmartvalues(&(data.vendor_attributes[i]), WORST_VALUE, 0 ) :
	     data.vendor_attributes[i].worse_normal),
	    /* JUNOS end */
	    thresholds.thres_entries[i].normalized_threshold);
	    for (j = 5 ; j >= 0 ; j--) {
		printf ("%02x",
		/* JUNOS begin */       	
	        (isunigenattributetbl() ?
	         printsmartvalues(&(data.vendor_attributes[i]), RAW_VALUE, j ) :
	         data.vendor_attributes[i].raw[j]));
		/* JUNOS end */
	    }
	    printf ("\n");
	}
    }
}

/* 
 * Prints verbose general smart values
 */
void 
ataprintgeneralsmartvalues  (struct ata_smart_values data)
{
    printf ("\nGeneral Smart Values: \n");

    printsmartofflinestatus (data); 
    printf("\n");
    
    if (issupportselftest(data)) {
        printsmartselfexecstatus (data);
        printf("\n");
    }
    
    printsmarttotaltimecompleteoffline (data);
    printf("\n");
    
    printsmartofflinecollectcap (data);
    printf("\n");
    
    printsmartcapability (data);
    printf("\n");

    printsmarterrorlogcapability (data);
    printf ("\n");
    
    if (issupportselftest(data)) {
        printsmartshortselftestpollingtime (data);
        printf ("\n");

        printsmartextendedselftestpollingtime (data);
        printf ("\n");
    }
}

/* 
 * Prints SMART thresholds
 */
static void __unused
ataprintsmartthresholds (struct ata_smart_thresholds data)
{
    int i;

    printf ("Smart Thresholds\n");
    printf ("Smart Threshold Revision Number: %i\n", data.revnumber);
    
    for (i = 0 ; i < NUMBER_ATA_SMART_ATTRIBUTES ; i++) {
        if (data.thres_entries[i].id != 0)    
            printf ("Atrribute %3i threshold: %02x (%2i)\n", 
	        data.thres_entries[i].id, 
                data.thres_entries[i].normalized_threshold, 
                data.thres_entries[i].normalized_threshold);
    }
}

/* 
 * Prints verbose SMART error log
 */
void 
ataprintsmarterrorlog (struct ata_smart_errorlog data)
{
    int i,j;

    printf ("\nSMART Error Log:\n");
    printf ("SMART Error Logging Version: %i\n", data.revnumber);

    if (! data.error_log_pointer) {
        printf ("No Errors Logged\n");
        return;
    }

    printf ("Error Log Data Structure Pointer: %02x\n", data.error_log_pointer);
    printf ("ATA Error Count: %u\n", data.ata_error_count);
    printf ("Non-Fatal Count: %u\n", data.non_fatal_count);
    
    for (i = 0; i < 5; i++) {
        if ((data.errorlog_struct[i].commands[0].devicecontrolreg ||
             data.errorlog_struct[i].commands[0].featuresreg ||
             data.errorlog_struct[i].commands[0].sector_count ||
             data.errorlog_struct[i].commands[0].sector_number ||
             data.errorlog_struct[i].commands[0].cylinder_low ||
             data.errorlog_struct[i].commands[0].cylinder_high ||
             data.errorlog_struct[i].commands[0].drive_head ||
             data.errorlog_struct[i].commands[0].commandreg) != 0) {

            printf("\nError Log Structure %i:\n", i+1);
            printf("DCR   FR   SC   SN   CL   SH   D/H   CR   Timestamp\n");
        
            for (j = 0; j < 6; j++) {
                printf (" %02x   %02x   %02x   %02x   %02x   %02x    %02x   %02x     %u\n", 
                    data.errorlog_struct[i].commands[j].devicecontrolreg,
                    data.errorlog_struct[i].commands[j].featuresreg,
                    data.errorlog_struct[i].commands[j].sector_count,
                    data.errorlog_struct[i].commands[j].sector_number,
                    data.errorlog_struct[i].commands[j].cylinder_low,
                    data.errorlog_struct[i].commands[j].cylinder_high,
                    data.errorlog_struct[i].commands[j].drive_head,
                    data.errorlog_struct[i].commands[j].commandreg,
                    (unsigned int) data.errorlog_struct[i].commands[j].timestamp / 1000); 
            }
        }

        if ((data.errorlog_struct[i].error_struct.error_condition ||
             data.errorlog_struct[i].error_struct.state) != 0) {

            printf ("Error condition: %3i\t",
            	data.errorlog_struct[i].error_struct.error_condition);
            printf ("Error State:     %3i\n", 
            	data.errorlog_struct[i].error_struct.state);
            printf ("Number of Hours in Drive Life: %u (life of the drive in hours)\n",  
                data.errorlog_struct[i].error_struct.timestamp);    
        } 
    }
}

char selftesttype[][14] =
{
    "Short        ",
    "Extended     ",
    "Short Cap    ",
    "Extended Cap ",
    "Unknown      "
};

char selfteststatus[][35] =
{
    "Successful                        ",
    "Aborted                           ",
    "Interrupted                       ",
    "Fatal error                       ",
    "Unknown failure                   ",
    "Electrical element failure        ",
    "Servo and/or seek element failure ",
    "Read element failure              ",
    "Reserved                          ",
    "Reserved                          ",
    "Reserved                          ",
    "Reserved                          ",
    "Reserved                          ",
    "Reserved                          ",
    "Reserved                          ",
    "In progress                       ",
    "Unknown                           "
};

/* 
 * Prints verbose SMART self test log
 */
void 
ataprintsmartselftestlog (struct ata_smart_selftestlog data) 
{
    int stat, i = 0;
    char *status;
    char *test_type;

    printf ("\nSMART SelfTest Log:\n");
    printf ("SMART SelfTest Logging Version: %i\n", data.revnumber);

    if (! data.mostrecenttest) {
        printf ("SelfTest has never been run\n");
        return;
    }


    printf("\nSelftest Type  Status                             Failure-LBA"
	    "  Timestamp\n\n");
    for (i = data.mostrecenttest - 1; i >= 0; i--) {

    	stat = data.selftest_struct[i].selfteststatus >> 4;
	stat = (stat <= 15) ? stat: 16;
	status = &selfteststatus[stat][0]; 

	switch(data.selftest_struct[i].selftestnumber) {
	    case SHORT_SELF_TEST:
		test_type = &selftesttype[0][0];
		break;
	    case EXTEND_SELF_TEST:
		test_type = &selftesttype[1][0];
		break;
	    case SHORT_CAPTIVE_SELF_TEST:
		test_type = &selftesttype[2][0];
		break;
	    case EXTEND_CAPTIVE_SELF_TEST:
		test_type = &selftesttype[3][0];
		break;
	    default:
		test_type = &selftesttype[4][0];
		break;
	}

	if (data.selftest_struct[i].lbafirstfailure) {
	    printf ("%s  %s 0x%08x   %u\n", test_type, status, 
		    data.selftest_struct[i].lbafirstfailure,
		    data.selftest_struct[i].timestamp);
	} else {
	    printf ("%s  %s None         %u\n", test_type, status, 
		    data.selftest_struct[i].timestamp);
	}
    }
}

/* 
 * Check SMART status by comparing values with thresholds
 */
void 
atapsuedochecksmart (struct ata_smart_values data, struct ata_smart_thresholds thresholds)
{
    int i;
    int failed = 0;

    for (i = 0 ; i < NUMBER_ATA_SMART_ATTRIBUTES ; i++) {
        if ((data.vendor_attributes[i].id !=0) &&   
            (thresholds.thres_entries[i].id != 0) &&
            (data.vendor_attributes[i].status_flag & 0x1) &&
            (data.vendor_attributes[i].normalized <
             thresholds.thres_entries[i].normalized_threshold) &&
            (thresholds.thres_entries[i].normalized_threshold != 0xFE)) {

           printf("Attribute %s, Failed\n", 
		  atagetsmartattribname(data.vendor_attributes[i].id));
           failed = 1;
        }
    }

#ifdef DEBUG_FAILURE
    printf("Attribute %s, Failed\n", atagetsmartattribname(200));
    failed = 1;
#endif

    if (failed) {
        printf ("Please save all data and call Juniper Technical Assistance "
		"Center immediately.\n");
    } else {
        printf ("Check SMART Passed\n");
    }
}

#define ATTR_SIZE 30
static char attribute_name[ATTR_SIZE];

/* 
 * Translate attribute id to attribute name
 */
char * 
atagetsmartattribname (unsigned char id)
{
    /* JUNOS begin */
    if (isvendorattribexists())
	return atagetvendorattribname(id);	
    /* JUNOS end */

    if (id >= 201) {
	snprintf(attribute_name, ATTR_SIZE, "(%3d)Vendor Unique          ", id);
	return attribute_name;
    }

    switch (id) {
    
	case 1:
	   snprintf(attribute_name, ATTR_SIZE, "(  1)Raw Read Error Rate    ");
	   break;
	case 2:
	   snprintf(attribute_name, ATTR_SIZE, "(  2)Throughput Performance ");
	   break;
	case 3:
	   snprintf(attribute_name, ATTR_SIZE, "(  3)Spin Up Time           ");
	   break;
	case 4:
	   snprintf(attribute_name, ATTR_SIZE, "(  4)Start Stop Count       ");
	   break;
	case 5:
	   snprintf(attribute_name, ATTR_SIZE, "(  5)Reallocated Sector Ct. ");
	   break;
	case 7:
	   snprintf(attribute_name, ATTR_SIZE, "(  7)Seek Error Rate        ");
	     break;
	case 8:
	   snprintf(attribute_name, ATTR_SIZE, "(  8)Seek Time Preformance  ");
	   break;
	case 9:
	   snprintf(attribute_name, ATTR_SIZE, "(  9)Power On Hours Count   ");
	    break;
	case 10:
	   snprintf(attribute_name, ATTR_SIZE, "( 10)Spin Retry Count       ");
	   break;
	  case 12:
	   snprintf(attribute_name, ATTR_SIZE, "( 12)Power Cycle Count      ");
	   break;
	 case 191:
	   snprintf(attribute_name, ATTR_SIZE, "(191)Gsense Error Rate      ");
	   break;
	 case 192:
	   snprintf(attribute_name, ATTR_SIZE, "(192)Power Off Retract Count");
	   break;
	 case 193:
	   snprintf(attribute_name, ATTR_SIZE, "(193)Load/Unload Cycle Count");
	   break;
	 case 194:
	   snprintf(attribute_name, ATTR_SIZE, "(194)Device Temperature     ");
	   break;
	 case 196:
	   snprintf(attribute_name, ATTR_SIZE, "(196)Reallocation Event Ct. ");
	   break;
	 case 197:
	   snprintf(attribute_name, ATTR_SIZE, "(197)Current Pending Sec. Ct");
	   break;
	case 198:
	   snprintf(attribute_name, ATTR_SIZE, "(198)Offline Uncorrectable  ");
	   break;
	case 199:
	   snprintf(attribute_name, ATTR_SIZE, "(199)UDMA CRC Error Count   ");
	   break;
	case 200:
	   snprintf(attribute_name, ATTR_SIZE, "(200)Write Error Rate       ");
	   break;
	default:
	   snprintf(attribute_name, ATTR_SIZE, "(%3d)Unknown Attribute      ", id);
	   break;
    }
    return attribute_name;
}    

int got_sigchld = 0;

void
atacaptivefinished(int signo)
{
    got_sigchld = signo;
}

#define MAXWAIT 60

static int 
ataprintdotsandcheck (int waittime)
{
    int i, pid, status, error;

    setbuf(stdout, NULL);
    /* 
     * Wait a while to see if the child aborts, in case the command is
     * not supported.
     */
    sleep(5);
    if (got_sigchld != 0) {
	return -1;
    }

    printf("\nWaiting for completion (est. %u, max. %d minutes) .", waittime,
	    MAXWAIT);
    
    waittime = MAXWAIT;

    for (i = 0; i < waittime ; i++) {
	if (got_sigchld != 0) {
	    if ((pid = wait3(&status, WNOHANG, (struct rusage *)NULL)) > 0) {
		if ((error = WEXITSTATUS(status)) == 0) {
		    printf("\nCommand complete after %d minutes.\n", i);
		    return 0;
		} else {
		    printf("\nChild exited with status %d\n", error); 
		    return -1;
		}
	    }
	    return -1;
	}
    	sleep(60);
	printf(".");
    }
    printf("\n%s: Exiting after the maximum wait time of %d minutes\n", 
		SMARTD, waittime);
    return -1;
}

static int
istestinprogress ( struct ata_smart_values smartval )
{
    if ((smartval.self_test_exec_status & 0xf0 ) == 0xf0 ) {
        printf ("%1d0%% of previously running test remaining\n", 
		smartval.self_test_exec_status & 0x0f);
	printf ("Please wait, or, use %s -%s to abort test\n", SMARTD, "oA");
	return TRUE;
    } else {
	return FALSE;
    }
}

/* 
 * Main routine to handle non daemon options
 */
int 
ataprintmain (int fd)
{
    struct ata_params drive;
    struct ata_smart_values smartval;
    struct ata_smart_thresholds smartthres;
    struct ata_smart_errorlog smarterror;
    struct ata_smart_selftestlog smartselftest;
    sigset_t omask, mask;
    int ret;

    sigemptyset(&mask);
    sigaddset(&mask, SIGINT);
    sigaddset(&mask, SIGQUIT);
    sigaddset(&mask, SIGHUP);

    if (formatunit) {
        if ( atareadhdidentity (fd, &drive) != 0) {
            printf("%s: Hard Drive Identity Failed\n", SMARTD);
            return -1;
        }
#ifdef DEBUG
	if ((drive.featsupp2 != 0xffff) && (drive.featsupp2 != 0x0000)) {
	    printf("featsupp1 = 0x%4x featsupp2 = 0x%4x\n", drive.featsupp1,
		   drive.featsupp2);
	    printf("featenab1 = 0x%4x featenab2 = 0x%4x\n", drive.featenab1,
		   drive.featenab2);
	}
#endif

	(void)signal(SIGCHLD, atacaptivefinished);
	sigprocmask(SIG_BLOCK, &mask, &omask);
	switch (fork()) {
	case 0: /* child */
	    if ((ret = ataformatunit(fd, &drive)) != 0) {
		if (ret == EOPNOTSUPP) {
		   printf("Disk model, %s, doesn't support the format command\n",
			    drive.model);
		}
		exit(1);	
	    }
	    exit(0);
	case -1:
            printf("%s: Could not fork (errno %d).\n", SMARTD, errno);
	    return -1;
	default:
	    /* erasetime is specified in units of 2 minutes */
	    return (ataprintdotsandcheck(drive.erase_time * 2));
	}
    }

    if (driveinfo) {
        if ( atareadhdidentity (fd, &drive) != 0) {
            printf("%s: Hard Drive Identity Failed\n", SMARTD);
            return -1;
        }
        ataprintdriveinfo(&drive); 
#ifdef DEBUG
	if ((drive.featsupp2 != 0xffff) && (drive.featsupp2 != 0x0000)) {
	    printf("featsupp1 = 0x%4x featsupp2 = 0x%4x\n", drive.featsupp1,
		   drive.featsupp2);
	    printf("featenab1 = 0x%4x featenab2 = 0x%4x\n", drive.featenab1,
		   drive.featenab2);
	}
#endif

        switch (atasmartsupport(&drive)) {
            case -1:
                printf ("Drive does not support SMART\n");
                return 0;
            case 1:
                printf ("Drive supports SMART and is disabled\n");
                printf("Use option -%s to enable\n","oe" );
                return 0;
            case 2: 
                printf ("Drive supports SMART and is enabled\n");
                break;
            case 255:
                printf ("Drive supports SMART and is ");
                if (atasmartstatus(fd) != 0) {
                    printf("disabled\n");
                    printf("Use option -%s to enable\n", "oe");
                    return 0;
                } else {
                    printf("enabled\n");
                }
                break;
            default :
                /* not used */
                break;
        }
    }

    if (smartdisable) {
        if ((ret = atadisablesmart(fd)) != 0) {
            if (ret == ENODEV) {
                printf ("Device does not support SMART, or SMART is already"
                                " disabled.\n");
                return 0;
            } else {
                printf("%s: Smart Disable Failed\n", SMARTD);
                return -1;
            }
        }
        printf("SMART Disabled\n");
        return 0;
    }

    if (smartenable) {
        if ((ret = ataenablesmart(fd)) != 0) {
            if (ret == ENODEV) {
                printf ("Device does not support SMART.\n");
                return 0;
            } else {
                printf("%s: Smart Enable Failed\n", SMARTD);
                return -1;
            }
        }
        printf("SMART Enabled\n");
        return 0;
    }
    
    /* For everything else reading of values and thresholds is needed */    

    if ((ret = atareadsmartvalues (fd, &smartval)) != 0) {
        if (ret == ENODEV) {
            printf ("Device does not support SMART, or SMART is disabled.\n");
            return 0;
        } else {
            printf("%s: Smart Values Read Failed\n", SMARTD);
            return -1;
        }
    }

    if (atareadsmartthresholds (fd, &smartthres) != 0) {
        printf("%s: Smart THresholds Read Failed\n", SMARTD);
    }
    
    if (checksmart) {
        /* pseudo is used because FreeBSD does not support access to
           Task File registers */
        atapsuedochecksmart (smartval , smartthres);
    }
    
    if ( generalsmartvalues) {
        ataprintgeneralsmartvalues(smartval); 
    }

    if (smartvendorattrib) {
        printsmartattribwiththres(smartval, smartthres);
    }

    if (smarterrorlog) {
        if (issmarterrorlogcapable(smartval) == 0) {
            printf("Device does not support Error Logging\n");
        } else {
            if (atareaderrorlog (fd, &smarterror) != 0) {
                printf("%s: Smart Errorlog Read Failed\n", SMARTD);
            } else {
                ataprintsmarterrorlog (smarterror); 
            }
        }
    }

    if (smartselftestlog) {
        if (issmarterrorlogcapable(smartval) == 0) {
            printf("Device does not support Self Test Logging\n");
        } else {
            if ( atareadselftestlog(fd, &smartselftest) != 0) {
                printf("%s: Smart Self Test log Read Failed\n", SMARTD);
            } else {
                ataprintsmartselftestlog (smartselftest); 
            }
        }
    }


    if ( smartautoofflineenable ) {
        if (!issupportautomatictimer (smartval)) {
            printf("Device does not support SMART Automatic Timers\n");
            return -1;
        }

        if (ataenableautooffline (fd) != 0) {
            printf("%s: Smart Enable Automatic Offline Failed\n", SMARTD);
            return -1;
        }
        printf ("SMART Automatic Offline Testing Enabled every four hours\n");
        return 0;
    }

    if ( smartautoofflinedisable ) {
        if (!issupportautomatictimer (smartval)) {
            printf("Device does not support SMART Automatic Timers\n");
            return -1;
        }
            
        if (atadisableautooffline (fd) != 0) {
            printf("%s: Smart Disable Automatic Offline Failed\n", SMARTD);
            return -1;
        }
        printf ("SMART Automatic Offline Testing Disabled\n");
        return 0;
    }

    if (smartexeoffimmediate) {
	if (istestinprogress(smartval)) {
	    return 0;
	}
	if (atasmartofflinetest (fd) != 0) {
            printf("%s: Smart Offline Failed\n", SMARTD);
            return -1;
        }
        printf ("Drive Command Successful, offline test has begun\n");
        printf ("Please wait %d seconds for test to complete\n",
                isofflinetesttime(smartval));

        printf ("Warning: Issuing any command to device, before test"
		" completion, will abort test.\n");
        return 0;
    }
    
    if (smartshortcapselftest) {
        if (! issupportselftest(smartval)) {
            printf (" ERROR: device does not support Self-Test function\n");
            return -1;
        }
	if (istestinprogress(smartval)) {
	    return 0;
	}

	(void)signal(SIGCHLD, atacaptivefinished);
	sigprocmask(SIG_BLOCK, &mask, &omask);

	switch (fork()) {
	case 0: /* child */
	    if (atasmartshortcapselftest (fd) != 0) {
		printf("%s: Smart Short Self Test Failed\n", SMARTD);
		exit(1);	
	    }
	    exit(0);
	case -1:
            printf("%s: Could not fork (errno %d).\n", SMARTD, errno);
	    return -1;
	default:
	    printf ("Drive Command Successful, captive short self test has begun\n");
	    printf ("Please wait %d minutes for test to complete\n", 
		    isshortselftesttime (smartval));
	    if (ataprintdotsandcheck (isshortselftesttime(smartval)) == 0) {
		if (atareadsmartvalues (fd, &smartval) != 0) {
		    printf("%s: Smart Values Read Failed\n", SMARTD);
		    return -1;
		}
		printsmartselfexecstatus (smartval);
		return 0;
	    } else {
		return -1;
	    }
	}
    }

    if (smartshortselftest) {
        if (! issupportselftest(smartval)) {
            printf (" ERROR: device does not support Self-Test function\n");
            return -1;
        }
	if (istestinprogress(smartval)) {
	    return 0;
	}
        
        if (atasmartshortselftest (fd) != 0) {
            printf("%s: Smart Short Self Test Failed\n", SMARTD);
            return -1;
        }
        printf ("Drive Command Successful, Short Self test has begun\n");
        printf ("Please wait %d minutes for test to complete\n", 
		isshortselftesttime (smartval));
        
        printf ("Use %s -%s to abort test\n", SMARTD, "oA");    
	return 0;
    }
    
    if (smartextendselftest) {
        if (! issupportselftest(smartval)) {
            printf (" ERROR: device does not support Self-Test function\n");
            return -1;
        }
	if (istestinprogress(smartval)) {
	    return 0;
	}

        if (atasmartextendselftest (fd) != 0) {
            printf("SMART Extendend Self Test Failed\n");
            return -1;
        }
        
        printf ("Drive Command Successful, Extended Self test has begun\n");
        printf ("Please wait %d minutes for test to complete\n", 
		isextendedselftesttime(smartval));
        
        printf ("Use %s -%s to abort test\n", SMARTD, "oA");    
        return 0;
    }

    if (smartextendcapselftest) {
        
        if (! issupportselftest(smartval)) {
            printf (" ERROR: device does not support self test function\n");
            return -1;
        }
	if (istestinprogress(smartval)) {
	    return 0;
	}

	(void)signal(SIGCHLD, atacaptivefinished);
	sigprocmask(SIG_BLOCK, &mask, &omask);

	switch (fork()) {
	case 0: /* child */
	    if (atasmartextendcapselftest (fd) != 0) {
		printf("SMART Extendend Self Test Failed\n");
		exit(1);	
	    }
	    exit(0);
	case -1:
            printf("%s: Could not fork (errno %d).\n", SMARTD, errno);
	    return -1;
	default:
	    printf ("Drive Command Successful, captive extended self test has begun\n");
	    printf ("Please wait %d minutes for test to complete\n", 
		    isextendedselftesttime (smartval));
	    if (ataprintdotsandcheck (isextendedselftesttime(smartval)) == 0) {
		if (atareadsmartvalues (fd, &smartval) != 0) {
		    printf("%s: Smart Values Read Failed\n", SMARTD);
		    return -1;
		}
		printsmartselfexecstatus (smartval);
		return 0;
	    } else {
		return -1;
	    }
	}
    }

    if (smartselftestabort) {

        if (! issupportselftest(smartval)) {
            printf ("ERROR: device does not support Self-Test function\n");
            return -1;
        }
	if ((smartval.self_test_exec_status & 0xf0 ) != 0xf0 ) {
	    printf ("No SMART selftest in progress. Nothing to abort.\n");
	    return 0;
	}

        if (atasmartselftestabort (fd) != 0) {
            printf("SMART Self Test Abort Failed\n");
            return -1;
        }
        printf ("Drive Command Successful, self test aborted\n");
        return 0;
    }        

    if (issuestandby) {
        if (ataissuestandbyimmediate (fd) != 0) {
            printf ("Could not issue standby immediate \n");
            return -1;
        }
        printf ("Successfully issued standby immediate \n");
        return 0;
    }
    return 0;
}
